/*    every.c    - generates a list of dates for 'every 2 weeks',
    'every 28 days' and so on. Will optionally generate mergeable
    lists though I'm not sure if this is really needed. Can input
    days, weeks or fortnights along with how many months worth
    (default 3).
*/

#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<ctype.h>
#include<time.h>
#include<unistd.h>


typedef struct tm TIME;
/*
struct tm {
    int tm_sec;             // Seconds.    [0-60] (1 leap second)
    int tm_min;             // Minutes.    [0-59]
    int tm_hour;            // Hours.    [0-23]
    int tm_mday;            // Day.        [1-31]
    int tm_mon;             // Month.    [0-11]
    int tm_year;            // Year - 1900.
    int tm_wday;            // Day of week.    [0-6]
    int tm_yday;            // Days in year.[0-365]
    int tm_isdst;           // DST.        [-1/0/1]
    long int __tm_gmtoff;   // Seconds east of UTC.
    __const char *__tm_zone;// Timezone abbreviation.
};
*/
typedef struct hm {
    int hh;         // [0..24]
    int mm;         // [0..59]
    int amflag;     // user spec'd am/pm time [0..1]
    int invalid;    // [0..1]
} HM;
typedef struct {
    int yy;         // year - as ccyy
    int mon;        // month
    int dd;         // day
    int hh;         // hour
    int mm;         // minute
    int wd;         // week day [0..6]
} DATEDATA;

void help(int err_status);
void arglist_print(int argc, char **argv);
HM *validate_time(const char *ts);
DATEDATA *validate_start(char *start);
void datedata2tm(TIME *tm, DATEDATA *dd);
void tm2datedata(DATEDATA *dd, TIME *tm);
void print_date_data(DATEDATA *tp, char **ddnames, char **mmnames,
                int reqtime, int mergeable, int ampm, int trunc_dm);


int main(int argc, char** argv) {
    int c;
    TIME *tp;
    time_t now, start, end;
    DATEDATA *dd;
    int hh, mm, print_time; // user specified time if any
    HM *hm;                 // ditto
    int amflag;             // 24 hr | am pm
    char *header;           // report header
    int inh_header;         // inhibit printing header
    int mergeable;          // output a leading "ccyymmddhhmm\t"
    // so that 2 or more runs may be meaningfully sorted eg events
    // happen at different times and start dates
    int trunc_dm;           // truncate day & month names to 3 chars
    int days;
    int months;
    char *start_date;
    char *mmnames[] = {"error",
                    "January",
                    "February",
                    "March",
                    "April",
                    "May",
                    "June",
                    "July",
                    "August",
                    "September",
                    "October",
                    "November",
                    "December"
    };
    char *ddnames[] = {"Sunday",
                    "Monday",
                    "Tuesday",
                    "Wednesday",
                    "Thursday",
                    "Friday",
                    "Saturday"
    };

    // set defaults
    print_time      = 0;    // don't print time unless user inputs it
    trunc_dm        = 0;    // print full month/day names unless -t spec'd
    hh              = 12;   // default time is noon
    mm              = 0;    // with 0 minutes after
    header          = strdup("Events"); // can be set by user
    inh_header      = 0;    // default is to print the header
    mergeable       = 0;    // default not to be mergeable
    days            = 0;    // error if not set to something
    months          = 3;    // may be set by user
    tp              = malloc(sizeof(TIME));
    while ((c = getopt(argc, argv, ":hc:d:w:fn:timl:")) != -1) {
        switch(c) {
            case 'h':
                help(0);
                break;
            case 'c':
                hm = validate_time(optarg);
                if (hm->invalid) {
                    printf("Invalid time %s\n", optarg);
                    arglist_print(argc, argv);
                    help(1);
                }
                amflag = hm->amflag;
                hh = hm->hh;
                mm = hm->mm;
                print_time = 1;
                free(hm);
                break;
            case 'd':
                days = atoi(optarg);
                if (days == 0){
                    fputs("Invalid no of days input\n", stderr);
                    arglist_print(argc, argv);
                    help(1);
                }
                break;
            case 'w':
                days = 7 * atoi(optarg);
                if (days == 0){
                    fputs("Invalid no of weeks input\n", stderr);
                    arglist_print(argc, argv);
                    help(1);
                }
                break;
            case 'n':
                months = atoi(optarg);
                if (months < 1){
                    fputs("One month minumum required\n", stderr);
                    arglist_print(argc, argv);
                    help(1);
                }
                break;
            case 'f':
                days = 14;
                break;
            case 't':
                trunc_dm = 1;
                break;
            case 'i':
                inh_header = 1;
                break;
            case 'm':
                mergeable = 1;
                break;
            case 'l':
                free(header);   // had default value
                header = strdup(optarg);
                break;
            case '?':
            case ':':
                fprintf(stderr,"Error in options:\n");
                arglist_print(argc, argv);
                if (c == '?')
                    fprintf(stderr,
                        "Illegal option supplied\n");
                else
                    fprintf(stderr,
                        "Option requires an argument\n");
                help(1);
                break;
            default:
                help(1);
                break;
        }// switch()
    }// while()
    // check that our options have been spec'd sanely
    if (days == 0){
        fputs("Options -d<days>, -w<weeks>, or -f required\n",stderr);
        arglist_print(argc, argv);
        help(1);
    }
    // should have 1 argv outside the options process, the start date
    start_date = strdup(argv[optind]);
    dd = validate_start(start_date);
    if (!(dd)) {
        fputs("Invalid starting date input\n", stderr);
        arglist_print(argc, argv);
        help(1);
    }
    // hours and minutes init to nonsense values so far, so:
    dd->hh  = hh;
    dd->mm  = mm;

    // out with the header if required
    if (!(inh_header)){
        if (mergeable)
            printf("000000000000\t%s\n", header);
        else
            printf("%s\n", header);
    }

    // set up start and end points
    datedata2tm(tp, dd);
    start       = mktime(tp);
    // deal with an end date possibly > 31-Dec-2037
    if (dd->yy == 2037 && months >= 12){
        tp->tm_year = 2037 - 1900;
        tp->tm_mon  = 11;   // [0..11]
        tp->tm_mday = 31 - days;
    } else {
        tp->tm_mon  += months;
    } // this lot needs more work, I'm falling 1 period short of
        // what I should let through but when I sail closer I roll
        // back to 1970. Will do for now though. I do not want to
        // make any test for this within the loop if I can avoid it.
    end         = mktime(tp);
    tp = localtime_r(&start, tp);
    for (now = start; now < end;) {
        tm2datedata(dd, tp);
        print_date_data(dd, ddnames, mmnames, print_time,
                        mergeable, amflag, trunc_dm);
        tp->tm_mday += days;
        now = mktime(tp);
        tp = localtime_r(&now, tp);
    }
    return 0;
}// main()
void help(int err_status){
    FILE *fpo;
    if (err_status)
        fpo = stderr;
    else
        fpo = stdout;
    fputs("NAME\n", fpo);
    fputs("\tevery.c - prints list of dates for regular events\n", fpo);
    fputs("SYNOPSIS\n", fpo);
    fputs("\tevery.c [OPTION]...StartDate[ccyymmdd]\n\n", fpo);
    fputs("DESCRIPTION\n", fpo);
    fputs("\tWrites to standard out.\n"
        "\tStarting with the input start date prints the dates of\n"
        "\tsuccessive periods for the nominated number of months\n"
        "\n"
    ,fpo);
    fputs("\t-h - prints this help\n\n", fpo);
    fprintf(fpo, "\t-d - Number of days apart, or\n");
    fprintf(fpo, "\t-w - number of weeks apart, or\n");
    fprintf(fpo, "\t-f - (no number) every fortnight.\n");
    fprintf(fpo, "\tIf use more than one of these options, only\n");
    fprintf(fpo, "\tthe last one seen will have any effect.\n\n");
    fprintf(fpo, "\t-n - Number of months to list, minimum 1.\n");
    fprintf(fpo, "\tThe default is 3 months.\n\n");

    fprintf(fpo, "\t-c - optional event time in 24 hr or am/pm format.\n");
    fprintf(fpo, "\tIf not specified no time will be printed.\n");
    fprintf(fpo, "\tIf time is input as am or pm it will be output\n");
    fprintf(fpo, "\tthat way, otherwise in 24 hour format.\n\n");
    fprintf(fpo, "\t-l - The headline for the report.\n");
    fprintf(fpo, "\tThe default is simply 'Events'\n\n");
    fprintf(fpo, "\t-i - Inhibit header printing entirely.\n");
    fprintf(fpo, "\tUse this when concatenating or merging a 2nd run.\n\n");
    fprintf(fpo, "\t-m - prepend a sort field so as to merge 2 or more\n");
    fprintf(fpo, "\truns. Sort the appended file, then cut -f2 to \n");
    fprintf(fpo, "\tget the merged result.\n\n");
    fprintf(fpo, "\t-t - Truncate the month and day names to 3\n");
    fprintf(fpo, "\tcharacters long for a columnar list.\n\n");
    exit(err_status);
}
void arglist_print(int argc, char **argv){
    int i;
    for (i = 1; i < argc; i++)
        fprintf(stderr, "%s  ", argv[i]);
    fputs("\n",stderr);
    return;
} // arglist_print()
HM *validate_time(const char *ts) {
    /* This returns the time in 24 hour format
        and a flag to indicate if user
        specified am or pm. Output time
        will be am or pm if user specified
        that way on input.
    */
    HM *hm;
    char *mp;
    char *buf = strdup(ts);
    char *work = strdup(ts);
    hm = (HM*) malloc(sizeof(HM));
    hm->invalid = 0;
    if (hm && buf && work) {
        hm->amflag = 0;
        if (strstr(buf, "am") ||
            strstr(buf, "AM") ||
            strstr(buf, "pm") ||
            strstr(buf, "PM") ){
            hm->amflag = 1;
        }
    // get hours/ minutes
    if ((mp = strstr(work, ":"))) {
        *mp = '\0';
        mp++;
        hm->hh = atoi(work);
        hm->mm = atoi(mp);
    } else if ((mp = strstr(work, "."))) {
        *mp = '\0';
        mp++;
        hm->hh = atoi(work);
        hm->mm = atoi(mp);
    } else {
        hm->hh = atoi(work);
        hm->mm = 0;
    }
    if (strstr(buf, "PM") || strstr(buf, "pm") ){
        if (hm->hh <= 12)
            hm->hh += 12;
        /* NB I am allowing the Thai syntax
        21.00 pm in this function */
    }
    if (hm->hh == 24) {
        hm->hh = 23;
        hm->mm = 59;
    }
    // validation
    if ( hm->hh < 0 || hm->hh >23 || hm->mm < 0 || hm->mm > 59 )
        hm->invalid = 1;
    } else {
        fprintf(stderr, "Could not get memory\n");
        exit(1);
    }
    return hm;
}// validate_time()
DATEDATA *validate_start(char *start) {
    // return the struct nicely filled in with our start date or NULL
    // start = ccyymmdd
    DATEDATA *tmp = malloc(sizeof(DATEDATA));
    tmp->hh = 12;   // fill it in for real later
    tmp->mm = 0;    // ditto
    tmp->wd = 0;    // ditto
    int len = strlen(start);
    if (!(len == 6 || len == 8))
        return NULL;
    tmp->dd = atoi(start + len - 2);
    if(tmp->dd < 1 || tmp->dd > 31)
        return NULL;
    start[len - 2] = '\0';
    tmp->mon = atoi(start + len - 4);
    if(tmp->mon < 1 || tmp->mon > 12)
        return NULL;
    start[len - 4] = '\0';
    tmp->yy = atoi(start);
    if (len == 6){
        if (tmp->yy < 36)
            tmp->yy += 2000;
        else if (tmp->yy >= 70)
            tmp->yy += 1900;
    }
    if (tmp->yy < 1970 || tmp->yy > 2037) // time_t rolls at those numbers
        return NULL;

    return tmp;
} // validate_start()
void datedata2tm(TIME *tm, DATEDATA *dd) {
    tm->tm_sec  = 0;
    tm->tm_min  = dd->mm;
    tm->tm_hour = dd->hh;
    tm->tm_mday = dd->dd;
    tm->tm_mon  = dd->mon - 1;
    tm->tm_year = dd->yy - 1900;
    return;
} // datedata2tm()
void tm2datedata(DATEDATA *dd, TIME *tm) {
    dd->yy  = tm->tm_year + 1900;
    dd->mon = tm->tm_mon + 1;
    dd->dd  = tm->tm_mday;
    dd->hh  = tm->tm_hour;
    dd->mm  = tm->tm_min;
    dd->wd  = tm->tm_wday;
    return;
} // tm2datedata()
void print_date_data(DATEDATA *tp, char **ddnames, char **mmnames,
        int reqtime, int mergeable, int ampm, int trunc_dm){
    char themon[10];
    char theday[10];
    strcpy(themon, mmnames[tp->mon]);
    strcpy(theday, ddnames[tp->wd]);
    if (trunc_dm) {
        themon[3] = '\0';
        theday[3] = '\0';
    }
    if (mergeable)
        printf("%4d%.2d%.2d%.2d%.2d\t",
               tp->yy,
               tp->mon,
               tp->dd,
               tp->hh,
               tp->mm );
    if (reqtime){
        if (!(ampm)){
            printf("%.2d:%.2d",
                   tp->hh,
                   tp->mm );
        } else {
            char *ap;
            if (tp->hh <= 12){
                ap = "AM";
            } else {
                tp->hh -= 12;
                ap = "PM";
            }
            printf("%2d:%.2d %s",
                   tp->hh,
                   tp->mm,
                   ap );
        }// if else(ampm)
    }
    printf(" %s  %2d-%s-%4d",
           theday,
           tp->dd,themon,
           tp->yy );
    puts("");   // at last a new line
    return;
}// print_date_data()
