/* xlist.c - a bundle of list processing functions
    These are general purpose functions every pointer being void.
    Hence the ordered insert function and both data based find and
    delete functions require the comparison function be specified, ie
    strcmp() for when data is of type char. For deletion the data destructor
    must be provided, and ie again for char* data this is simply free().
    There are 2 traversal functions, forward and reverse, and these require
    a function to process the data at the node. If it's char data and you
    want to print it, puts() does the job nicely.

    Version 0.1 - 2008-05-22
*/

#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<ctype.h>
#include "xlist.h"


LNODE *new_node(void){
	LNODE *tmp = malloc(sizeof(LNODE));
	if (tmp == NULL)
		fatal_error("No memory in new-node()");
	tmp->data = NULL;
	tmp->next = NULL;
	return tmp;
} // new_node()
void destroy_node(LNODE *node, void (*destroy)(void *)) {
	// destroy is the destructor for node->data
	// of course it may just be free()
	destroy(node->data);
        node->data = NULL;
	free(node);
        node = NULL;
	return;
} // destroy_node()
LNODE *ol_insert(LNODE *olnode, LNODE *toinsert, int (*comp)(void *, void *)){
	// traverse the list, if x->next->data is greater
	// or x->next is null; insert the new one here
	for (; olnode->next ; olnode = olnode->next) {
		if ((*comp)(toinsert->data, olnode->next->data) < 0)
			break; 
	} // for()
	// do the insert
	toinsert->next = olnode->next; // NULL or not? no matter
	olnode->next = toinsert;
	return toinsert;
} // ol_insert()
void node_traverse(LNODE *node, int (*do_what)(void *)) {
	// same process, ordered or unordered
	for(; node; node = node->next) {
		if (node->data)
                        (*do_what)(node->data);
	} // for()
	return;
} // ol_prints
void node_traverse_reverse(LNODE *node, int (*do_what)(void *)) {
	// recursion is so natural it's irresistable
	if (!(node))
		return;
	node_traverse_reverse(node->next, (void*)do_what);
	// the list head data is always NULL
	if (node->data)
                (*do_what)(node->data);
	return;
} // node_traverse_reverse()
void fatal_error(char *msg) {
	// but will perror() work when we can't malloc()????
	perror(msg);
	exit(1);
} // void fatal_error()
void destroy_list(LNODE *head, void (*destroy)(void *)) {
	// naturally recursive
	if (!(head))
		return;
	destroy_list(head->next, destroy);
	destroy_node(head, destroy);
	return;
} // destroy_list()
LNODE *ul_append(LNODE *head, LNODE *newnode) {
	// drops newnode at the end so its an
	// unordered list. Of course it may still
	// be ordered if the newnode is generated
	// in order.

	for (; head->next; head = head->next)
		;
	head->next = newnode;
	return newnode;
} // ul_append()
LNODE *ul_push(LNODE *head, LNODE *newnode) {
	newnode->next = head->next;
	head->next = newnode;
	return head;
} // ul_push()
LNODE *ul_pop(LNODE *head) {
	// the list head is a dummy so:
	LNODE *tmp = head->next; // 1st node in list
	if (head->next)
		head->next = head->next->next;
	return tmp;	// it may be NULL
} // ul_pop()
LNODE *ol_find(LNODE *olnode, void *what, int (*comp)(void *, void *)){
	// use find() to access node->data
	LNODE *tmp = ol_find_before(olnode, what, comp);
	if (tmp)
		return tmp->next;
	return NULL;
}// ol_find()
LNODE *ol_find_before(LNODE *olnode, void *what, int (*comp)(void *, void *)){
	// use find_before() when intending to delete the node
	// NB the list head is always a dummy node, so:
	int c;
	for (; olnode->next; olnode = olnode->next) {
		c = (*comp) (what, olnode->next->data);
		if (c == 0)
			return olnode;
		if ( c < 0)
			return NULL;
	} // for()
	return NULL;
}// ol_find_before()
LNODE *ul_find_before(LNODE *ulnode, void *what, int (*comp)(void *, void *)){
	// use find_before() when intending to delete the node
	// NB the list head is always a dummy node, so:
	int c;
	for (; ulnode->next; ulnode = ulnode->next) {
		c = (*comp) (what, ulnode->next->data);
		if (c == 0)
			return ulnode;
	} // for()
	return NULL;
}// ul_find_before()
LNODE *ul_find(LNODE *ulnode, void *what, int (*comp)(void *, void *)){
	// use find() to access node->data
	LNODE *tmp = ul_find_before(ulnode, what, comp);
	if (tmp)
		return tmp->next;
	return NULL;
}// ul_find()
LNODE *ul_delete(LNODE *node, void *what,
		int (*comp)(void *, void *), void (*data_destroy) (void)){
	// locate the node to delete, if it exists delete it and return
	// the pointer to the node before it, if not return NULL.
	LNODE *nodeb4, *del;
	nodeb4 = ul_find_before(node, what, (void*)comp);
	if (nodeb4) {
		del = nodeb4->next;
		nodeb4->next = nodeb4->next->next;
		destroy_node(del, (void*)data_destroy);
		return nodeb4;
	} // if()
	return NULL;
} // ul_delete()
LNODE *ol_delete(LNODE *node, void *what, 
		int (*comp)(void *, void *), void (*data_destroy) (void)){
	// locate the node to delete, if it exists delete it and return
	// the pointer to the node before it, if not return NULL.
	LNODE *nodeb4, *del;
	nodeb4 = ol_find_before(node, what, (void*)comp);
	if (nodeb4) {
		del = nodeb4->next;
		nodeb4->next = nodeb4->next->next;
		destroy_node(del, (void*)data_destroy);
		return nodeb4;
	} // if()
	return NULL;
} // ol_delete()
LNODE *node_delete(LNODE *node, LNODE *delete,
			void (*data_destroy) (void)){
	// The fact that we have a node to destroy implies that we 
	// don't know where it is, ie the node before it so:
	// locate this node, if it exists delete its ->next and
	// return the pointer to this, if not return NULL. 
	LNODE *del;
	for (; node->next; node = node->next)
		if (node->next == delete)
			break;
	if (node->next) {
		del = node->next;
		node->next = node->next->next;
		destroy_node(del, (void*)data_destroy);
		return node;
	} // if()
	return NULL;
} // ol_delete()

// TODO no doubt I can expand this to implement a queue.
// Actually ul_append() to put data in and ul_pop() to get it out
// acts as a FIFO.
